// Copyright 2019, the Chromium project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'dart:async';
import 'dart:convert';
import 'package:flutter/foundation.dart' show kIsWeb;
import 'package:flutter_driver/driver_extension.dart';
import 'package:flutter_test/flutter_test.dart';
// ignore: implementation_imports
import 'package:test_api/src/backend/group.dart' show Group;
// ignore: implementation_imports
import 'package:test_api/src/backend/group_entry.dart' show GroupEntry;
// ignore: implementation_imports
import 'package:test_api/src/backend/invoker.dart' show Invoker;

import 'common.dart';

typedef TestMain = void Function();

List<Map<String, dynamic>> _pendingEvents = [];

void main(testMain) {
  String currentTestId = '0';
  Map<Object, Map<String, String?>> testsMap = {};
  Map<Object, Map<String, String?>> groupMap = {};

  enableFlutterDriverExtension(handler: (data) async {
    if (_pendingEvents.isEmpty) return '';
    return json.encode(_pendingEvents.removeAt(0));
  });

  runZonedGuarded(() {
    setUp(() {
      Map<String, String?> currentTest =
          testsMap[Invoker.current!.liveTest.test]!;
      currentTestId = currentTest[kId]!;
      _pendingEvents.add({
        kAction: EventActionTestStart,
        kData: {
          kId: currentTest[kId],
          kName: currentTest[kName],
          kGroupId: currentTest[kGroupId],
          kGroupName: currentTest[kGroupName],
        }
      });
    });

    setUpAll(() {
      void iterateGroupEntries(GroupEntry groupEntry, GroupEntry parentEntry) {
        if (groupEntry is Group) {
          if (groupMap[groupEntry] == null) {
            Map<String, String?>? parentGroup = groupMap[parentEntry];

            groupMap[groupEntry] = {
              kId: (groupMap.length + 1).toString(),
              kName: groupEntry.name.replaceFirst(parentEntry.name, '').trim(),
              "parentGroupId": parentGroup != null ? parentGroup[kId] : null,
              "parentGroupName":
                  parentGroup != null ? parentGroup[kName] : null,
            };
          }

          groupEntry.entries.forEach((entry) {
            iterateGroupEntries(entry, groupEntry);
          });
        } else {
          if (testsMap[groupEntry] == null) {
            Map<String, String?> parentGroup = groupMap[parentEntry]!;

            testsMap[groupEntry] = {
              kId: (testsMap.length + 1).toString(),
              kName: groupEntry.name.replaceFirst(parentEntry.name, '').trim(),
              kGroupId: parentGroup[kId],
              kGroupName: parentGroup[kName],
            };
          }
        }
      }

      Invoker.current!.liveTest.groups.forEach((group) {
        if (groupMap[group] == null) {
          groupMap[group] = {
            kId: (groupMap.length + 1).toString(),
            kName: group.name,
          };
        }
        group.entries.forEach((entry) {
          iterateGroupEntries(entry, group);
        });
      });

      _pendingEvents.add({
        kAction: EventActionSuiteStart,
        kData: {
          'tests': testsMap.values.toList(),
          'groups': groupMap.values.toList(),
          kWeb: kIsWeb,
        }
      });
    });

    tearDownAll(() {
      currentTestId = '0';
      _pendingEvents.add({
        kAction: EventActionSuiteComplete,
      });
    });

    tearDown(() {
      Map<String, String?> currentTest =
          testsMap[Invoker.current!.liveTest.test]!;

      Map eventData = {
        kId: currentTest[kId],
        kName: currentTest[kName],
        kGroupId: currentTest[kGroupId],
        kGroupName: currentTest[kGroupName],
        'result': Invoker.current!.liveTest.state.result.toString(),
        'passed': Invoker.current!.liveTest.state.result.isPassing,
        kWeb: kIsWeb,
      };

      if (eventData['passed'] == false) {
        AsyncError errorForTest = Invoker.current!.liveTest
            .errors[Invoker.current!.liveTest.errors.length - 1];
        eventData[kError] = errorForTest.error.toString();
        eventData[kStacktrace] = errorForTest.stackTrace.toString();
      }

      _pendingEvents.add({
        kAction: EventActionTestComplete,
        kData: eventData,
      });
    });

    testMain();
  }, (e, stacktrace) {
    Map<String, String?>? currentTest =
        testsMap[Invoker.current?.liveTest.test ?? 1];
    _pendingEvents.add({
      kAction: EventActionError,
      kData: {
        kError: e.toString(),
        kStacktrace: stacktrace.toString(),
        kTestId: currentTest != null ? currentTest[kId] : null,
        kWeb: kIsWeb,
      },
    });
    _pendingEvents.add({
      kAction: EventActionSuiteComplete,
    });
  }, zoneSpecification: ZoneSpecification(
      print: (Zone self, ZoneDelegate parent, Zone zone, String message) {
    if (kIsWeb) {
      // TODO(salakar) skipping web stack trace logging for now until source-map support added.
      if (message.contains('main.dart.js')) {
        return;
      }
      // TODO(salakar) skipping reporter log lines for web
      if (message.startsWith(r'[0-9]{2}:[0-9]{2}\s\+.*')) {
        return;
      }
    } else {
      String stacktrace = StackTrace.current.toString();
      if (stacktrace.contains('_Reporter._progressLine') ||
          stacktrace.contains('_Reporter._onError')) return;
    }
    _pendingEvents.add({
      kAction: EventActionPrint,
      kData: {
        kMessage: message,
        kTestId: currentTestId,
      },
    });
  }));
}
